
from .grammar import EOF, Symbol
from .engine import Engine, State, Table, Token

class FSMTable (Table):

    def write_py(self, f, indent=''):
        f.write("{}from grammar.eng_fsm import FSMTable, State\n".format(indent))
        f.write("{}TABLE = FSMTable({!r}, {!s})\n".format(indent, self.name, self.nstart))
        f.write("{}TABLE.states = [\n".format(indent))
        for n,s in enumerate(self.states):
            f.write("{}    State('', {!r}, {!r}),\n".format(indent,s.action,n))
        f.write("{}    ]\n\n".format(indent))

    def write_etree(self):
        from xml.etree.ElementTree import Element, SubElement
        te = Element("fsm", name=self.name)
        te.text="\n"
        for n,s in enumerate(self.states):
            se = SubElement(te, "state", id=str(n))
            se.text="\n"
            se.tail="\n"
            for k in s.action:
                ds, act = s.action[k]
                d = { 'dest': str(ds) }
                if act: d['action'] = str(act)
                if isinstance(k, str):
                    ae = SubElement(se, "text", value=k, **d)
                else:
                    ae = k.etree(se, SubElement, d)
                ae.tail="\n"
        return te

class FiniteStateMachine (Engine):

    def get_state(self):
        return self.state

    def parse(self, st=0):
        self.state = self.TABLE.states[st]
        super().parse()

    def handle(self, s, a, t):
        if s is None or s == 'e':
            self.finished = True
        else:
            self.state = self.TABLE.states[s]
        self.action(a, t)

    def action(self, a, t, **kws):
        if a is None: return None
        elif isinstance(a, str):
            f = getattr(self, a)
        else:
            f = a
        return f(t, **kws)

class LexicalAnalyser:

    SPACE_TKN = object()

    def get(self):
        self.tkn = self.SPACE_TKN
        self.nlignore = False
        while self.tkn is self.SPACE_TKN:
            self.text = []
            self.tkn = None
            self.tsrc = None
            self.init_token()
            self.parse(int(self.linestart and self.TABLE.nstart == 2))
            if self.tkn is None:
                if not self.text:
                    if isinstance(self.t, Token) and self.t.type is EOF:
                        return self.t
                    elif self.t is EOF:
                        return Token('', EOF)
                raise self.error(self.t)
            else:
                t = self.text[self.tlen-1]
                if isinstance(t, Token): t = t.text
                self.linestart = (not self.nlignore and t == '\n')
            self.src.unget(self.text[self.tlen:])
        return self.tkn

    def ignore_newline(self):
        self.nlignore = True

    def init_parse(self):
        self.linestart = True

    def init_token(self):
        pass

    def action(self, a, t):
        if isinstance(t, Token):
            s = t.src
            tk = True
            eof = t.type is EOF
        else:
            s = None
            tk = False
            eof = t is EOF
        if not eof:
            if not self.text: self.tsrc = s
            self.text.append(t)
        if a is not None:
            if tk:
                text = ''.join(x.value for x in self.text)
            else:
                text = ''.join(self.text)
            if isinstance(a, str):
                f = getattr(self, a)
            else:
                f = a
            if s is None:
                tsrc = self.tsrc
            else:
                tsrc = self.tsrc|s
            if isinstance(f, Symbol):
                r = Token(text, f, src=tsrc)
            else:
                r = f(text, src=tsrc)
            if r is not None:
                self.tkn = r
                self.tlen = len(self.text)

    def SPACE(self, text, src):
        return self.SPACE_TKN

    LITERAL_TOKEN = Token
