
from .grammar import Symbol, EOF, Ref
from .engine import Engine, State, Table

class Rule:
    def __init__(self, lhs, nsyms, action, args, src=None):
        self.src = src
        self.lhs = lhs
        self.nsyms = nsyms
        self.action = action
        self.args = args

class ShiftReduceTable (Table):
    def __init__(self, name):
        super().__init__(name)
        self.rules = []
        self.pmap = {}

    def ensure_rule(self, p):
        pi = self.pmap.get(p)
        if pi is None:
            pi = len(self.rules)
            pp = Rule(p.lhs, len(p.rhs), p.action, p.args, p.src)
            pp.raw = p
            self.rules.append(pp)
            self.pmap[p] = pi
        return pi

    def add_goto(self, s, nt, d):
        self.add(s, nt, 'g', d)

    def add_accept(self, s, t):
        self.add(s, t, 'a', None)

    def add_shift(self, s, t, d):
        self.add(s, t, 's', d)

    def add_reduce(self, s, t, p):
        self.add(s, t, 'r', p)

    def write_py(self, f, indent=''):
        f.write("{}from grammar.eng_sr import ShiftReduceTable, State, Rule\n".format(indent))
        syms = { p.lhs for p in self.rules }
        for s in self.states:
            syms |= { x for x in s.action if isinstance(x, Symbol)}
        if EOF in syms: syms.remove(EOF)
        for s in syms:
            if s.anon:
                f.write("{0}{1} = NonTerminal({1!r})\n"
                        .format(indent, s.name))
            else:
                f.write("{0}{1} = {2}.{1}\n".format(indent,s.name, self.name))
        f.write("\n{}TABLE = ShiftReduceTable({!r})\n".format(indent,self.name))
        f.write("{}TABLE.states = [\n".format(indent))
        for n,s in enumerate(self.states):
            f.write("{}    #STATE {}\n".format(indent, s.n))
            f.write("{}    State({!r}, {!r}, {!r}),\n".format(indent, s.src, s.action, n))
        f.write("{}    ]\n\n".format(indent))
        f.write("{}TABLE.rules = [\n".format(indent))
        for p in self.rules:
            if isinstance(p.action, str):
                act = repr(p.action)
            elif isinstance(p.action, type):
                act = p.action.__name__
            elif p.action is None:
                act = "None"
            else:
                act = "<function>"
            f.write("{}    Rule({}, {}, {}, {!r}),\n"
                    .format(indent,p.lhs,p.nsyms,act,p.args))
        f.write("{}    ]\n\n".format(indent))

class LRStandardActions:

    @staticmethod
    def COPY(x, *, src=None):
        return x

    @staticmethod
    def TUPLE(*x, src=None):
        return x

    @staticmethod
    def DICT(src=None, **x):
        return x

    @staticmethod
    def LIST(x, *, src=None):
        return [x]

    @staticmethod
    def EMPTYLIST(*, src=None):
        return []

    @staticmethod
    def APPEND(l, x, *, src=None):
        return l+[x]

    @staticmethod
    def PREPEND(x, l, *, src=None):
        return [x]+l

class ShiftReduceEngine (Engine, LRStandardActions):
    
    def run(self, st=0):
        self.stack = [(self.TABLE.states[st], None)]
        self.parse()
        return self.stack[-1][1]

    def get_state(self):
        return self.stack[-1][0]

    def handle(self, c, a, t):
        while c == 'r':
            p = self.TABLE.rules[a]
            v = self.action(p, self.stack)
            if p.nsyms: self.stack[-p.nsyms:] = []
            g = self.stack[-1][0].action[p.lhs][1]
            self.stack.append((self.TABLE.states[g], v))
            c,a = self.match(t, self.stack[-1][0])
        if c == 'e':
            raise self.error(t)
        elif c == 's':
            self.stack.append((self.TABLE.states[a],t))
        elif c == 'a':
            self.finished = True
        else:
            assert False

    def action(self, p, argv):
        if isinstance(p.action, str):
            f = getattr(self, p.action)
        else:
            f = p.action
        kwargs = { k:(self.arg(argv[v-p.nsyms][1])
                      if isinstance(v, Ref) else v)
                   for k,v in p.args[1].items() if k != '!!' }
        for v in p.args[1].get('!!', []):
            kwargs.update(self.arg(argv[v-p.nsyms][1]))
        kwargs.update(self.extra(p, argv))
        args = ( (self.arg(argv[v-p.nsyms][1])
                  if isinstance(v, Ref) else v) for v in p.args[0] )
        return self.result(f(*args, **kwargs), p.lhs)

    def arg(self, v):
        return v

    def result(self, r, nt):
        return r

    def extra(self, p, argv):
        return {}
