
from .grammar import ANY, TYP, SET, TSYM, NTSYM, EOF, \
    NonTerminal, RMacro, Category

class ParseError (Exception):
    pass

class SemanticError (Exception):
    def __str__(self):
        if len(self.args) != 2: return super().__str__()
        if len(self.args[1]) == 0: return self.args[0]
        s = self.args[1][0].get_src()
        if s is None: return self.args[0]
        return self.args[0]+' at '+str(s)

class MultipleSemanticError (SemanticError):
    pass

class Token:
    def __init__(self, text, type=None, value=None, src=None):
        self.text = text
        self.type = text if type is None else type
        self.value = text if value is None else value
        self.src = src

    def get_src(self):
        return self.src

    def __repr__(self):
        return "<Token {}>".format(self.text)

class Ambiguity (list):

    def __getattr__(self, n):
        if len(self) == 1:
            return getattr(self[0], n)
        def _fn(*args, _self=self, _n=n, **kwds):
            nl = []
            errs = []
            rl = []
            for v in _self:
                try:
                    r = getattr(v, _n)(*args, **kwds)
                except MultipleSemanticError as e:
                    errs.extend(e.args)
                except SemanticError as e:
                    errs.append(e)
                else:
                    rl.append(r)
                    nl.append(v)
            _self[:] = nl
            if len(_self) == 1:
                return rl[0]
            elif _self:
                for i,v in enumerate(_self):
                    if hasattr(v, 'resolve_after'):
                        j = v.resolve_after(self, i)
                        if j is not None:
                            _self[:] = _self[j:j+1]
                            return rl[j]
                raise SemanticError("Unresolved Ambiguity", self[0].src)
            elif errs:
                raise MultipleSemanticError(errs)
            else:
                raise SemanticError("No available possibilities")
        return _fn

class ASTNode:
    def __init__(self, **children):
        for k in children:
            setattr(self, k, children[k])

    def get_self(self):
        return self

    def get_src(self):
        if self.src:
            return self.src[0].get_src()
        return None

    def display(self, f, gen=0):
        ind = '  '*gen
        f.write('{} (\n'.format(type(self).__name__))
        for k,v in self.__dict__.items():
            if k == 'src': continue
            f.write('{}  {} = '.format(ind, k))
            self._display_item(v, f, gen+1)
        f.write('{})\n'.format(ind))

    def deep_copy(self, idmap=None):
        if idmap is None: idmap = {}
        nd = {k:self._deep_copy(v, idmap) for k,v in self.__dict__.items()}
        return type(self)(**nd)

    @classmethod
    def _deep_copy(cls, v, idmap):
        if id(v) in idmap: return idmap[id(v)]
        if isinstance(v, ASTNode):
            nv = v.deep_copy(idmap)
        elif isinstance(v, Ambiguity):
            nv = Ambiguity(cls._deep_copy(x, idmap) for x in v)
        elif isinstance(v, list):
            nv = [cls._deep_copy(x, idmap) for x in v]
        elif isinstance(v, tuple):
            nv = tuple([cls._deep_copy(x, idmap) for x in v])
        else:
            return v
        idmap[id(v)] = nv
        return nv

    @classmethod
    def _display_item(cls, v, f, gen):
        ind = '  '*gen
        if isinstance(v, Ambiguity):
            if len(v) == 1:
                cls._display_item(v[0], f, gen)
            else:
                f.write('<<<\n')
                for i in v:
                    f.write('{}  '.format(ind))
                    cls._display_item(i, f, gen+1)
                f.write('{}>>>\n'.format(ind))
        elif isinstance(v, list):
            f.write('[[[\n')
            for i in v:
                f.write('{}  '.format(ind))
                cls._display_item(i, f, gen+1)
            f.write('{}]]]\n'.format(ind))
        elif isinstance(v, ASTNode):
            v.display(f, gen)
        else:
            f.write('{}\n'.format(v))

class Buffer:
    """Buffer between source that can provide one or more symbols at a time
    and a receiver that accepts only one at a time, and may go back"""

    def __init__(self, src):
        self.src = src
        self.queue = []
        self.ended = False

    def get(self):
        "Get a single symbol"
        if len(self.queue) == 0:
            self._fill()
            if len(self.queue) == 0:
                return EOF()
        return self.queue.pop(0)

    def unget(self, o):
        "Unget one or more symbols.  If list: in order they were received"
        if isinstance(o, list):
            self.queue = o + self.queue
        else:
            self.queue.insert(0, o)

    def _fill(self):
        o = self.src.get()
        if isinstance(o, list):
            self.queue += o
        else:
            self.queue.append(o)

    def filename(self):
        return self.src.filename()

class State:
    def __init__(self, src, action=None, n=None):
        self.src = src
        self.n = n
        if action:
            self.action = action
        else:
            self.action = {}

    def __repr__(self):
        return "<STATE {}>".format(self.n)

class Table:
    def __init__(self, name, nstart=1):
        self.name = name
        self.nstart = nstart
        self.states = []
        self.smap = {}
        self.conflicts = []

    def ensure_state(self, s, gen):
        si = self.smap.get(s)
        if si is None:
            si = len(self.states)
            st = State([x.source() for x in gen.state_items(s)])
            st.raw = s
            st.n = si
            self.states.append(st)
            self.smap[s] = si
        return si

    def add(self, s, tt, c, a):
        st = self.states[s]
        if isinstance(tt, SET):
            if not tt.inv:
                for t in tt.s:
                    self.add(s, TSYM(t), c, a)
                return
            t = ANY()
        else:
            t = tt.match()
        """
        if isinstance(t, ANY) and \
                not all(isinstance(x, NTSYM) or
                        (isinstance(x, TSYM) and x.v is EOF)
                        for x in st.action):
            cc,ca = self.states[s].action.get(t, ('d', None))
        elif isinstance(t, TYP) and \
                any(isinstance(x, TSYM) and category(x.v) == t.v
                    for x in st.action):
            cc,ca = self.states[s].action.get(t, ('d', None))
        el"""
        if t not in st.action:
            self.states[s].action[t] = (c,a)
            return
        else:
            cc,ca = self.states[s].action[t]
        if cc == 'c':
            ca.append((c,a))
            return
        self.states[s].action[t] = ('c', [(cc, ca), (c, a)])
        self.conflicts.append((s,t))

class Engine:
    def __init__(self, src, *args, **kwargs):
        self.src = Buffer(src)
        self.init_parse(*args, **kwargs)

    def parse(self):
        self.finished = False
        while not self.finished:
            t = self.src.get()
            self.t = t
            m = self.match(t, self.get_state())
            self.handle(*m, t=t)

    def match(self, t, s):
        (tc, tt, ta) = self.match_with(t)
        try:
            return s.action[tc]
        except KeyError:
            try:
                return s.action[TYP(tt)]
            except KeyError:
                if ta:
                    return s.action.get(ANY(), ('e', None))
                else:
                    return ('e', None)

    def match_with(self, t):
        return (t, (self.category(t) if t is not EOF else None), t is not EOF)

    def init_parse(self, *args, **kwargs):
        pass

    def category(self, t):
        return None

class MatchToken:
    def match_with(self, t):
        return (t.type, (self.category(t.text) if t.text else None),
                t.type is not EOF)

    def arg(self, v):
        return v.value

    def result(self, r, nt):
        return Token(None, nt, r)

    def run(self, st=0):
        r = super().run(st)
        return r.value

class BlankSourceTrack:
    def __or__(self, other):
        return other

class SourceTrace (MatchToken):

    def extra(self, p, argv):
        if p.nsyms == 0:
            fst = BlankSourceTrack()
        else:
            fst = argv[-p.nsyms][1].src | argv[-1][1].src
        self.x_fst = fst
        return { 'src':fst }

    def result(self, r, nt):
        return Token(None, nt, r, src=self.x_fst)

    def error(self, tkn):
        raise ParseError("{}: Unexpected {!r} at {!s}".format(type(self).__name__, tkn.text, tkn.src))

class Reproduce (SourceTrace):
    def extra(self, p, argv):
        if p.nsyms == 0:
            pt = []
        else:
            pt = [a[1] for a in argv[-p.nsyms:]]
        self.x_fst = pt
        return { 'src': pt }

    def result(self, r, nt):
        if not isinstance(r, ASTNode):
            return Token(None, nt, r, src=self.x_fst)
        elif r.src is self.x_fst:
            src = []
            for s in self.x_fst:
                self.flatten_src(src, s)
            r.src = src
            return Token(None, nt, r, src=[r])
        else:
            return Token(None, nt, r, src=self.x_fst)

    def flatten_src(self, r, s):
        if isinstance(s, ASTNode):
            r.append(s)
        elif isinstance(s, Ambiguity):
            r.append(s)
        elif isinstance(s.src, list):
            for n in s.src:
                self.flatten_src(r, n)
        else:
            r.append(s)

class UnicodeCategories:
    C = RMacro("({Cc}|{Cf}|{Cn}|{Co}|{Cs})")
    L = RMacro("({Ll}|{Lm}|{Lo}|{Lt}|{Lu})")
    M = RMacro("({Mc}|{Me}|{Mn})")
    N = RMacro("({Nd}|{Nl}|{No})")
    P = RMacro("({Pc}|{Pd}|{Pe}|{Pf}|{Pi}|{Po}|{Ps})")
    S = RMacro("({Sc}|{Sk}|{Sm}|{So})")
    Z = RMacro("({Zl}|{Zp}|{Zs})")
    Other_ID_Start = RMacro("[\u2118\u212e\u309b\u309c]")
    Other_ID_Continue = RMacro("[\xb7\u0387\u1369-\u1371\u19da]")
    White_Space = RMacro("[\x09-\x0d \x8c\xa0\u1680\u180e\u2000-\u200a\u2028"
                         "\u2029\u202f\u205f\u3000]")
    Hex_Digit = RMacro("[0-9A-Fa-f\uff10-\uff19\uff21-\uff26\uff41-\uff46]")
    
    Cc = Category()
    Cf = Category()
    Cn = Category()
    Co = Category()
    Cs = Category()
    Ll = Category()
    Lm = Category()
    Lo = Category()
    Lt = Category()
    Lu = Category()
    Mc = Category()
    Me = Category()
    Mn = Category()
    Nd = Category()
    Nl = Category()
    No = Category()
    Pc = Category()
    Pd = Category()
    Pe = Category()
    Pf = Category()
    Pi = Category()
    Po = Category()
    Ps = Category()
    Sc = Category()
    Sk = Category()
    Sm = Category()
    So = Category()
    Zl = Category()
    Zp = Category()
    Zs = Category()

    from unicodedata import category

    HEXADECIMAL_VALUE = {
        '0': 0,
        '1': 1,
        '2': 2,
        '3': 3,
        '4': 4,
        '5': 5,
        '6': 6,
        '7': 7,
        '8': 8,
        '9': 9,
        'A': 10,
        'B': 11,
        'C': 12,
        'D': 13,
        'E': 14,
        'F': 15,
        'a': 10,
        'b': 11,
        'c': 12,
        'd': 13,
        'e': 14,
        'f': 15,
        '\uff10': 0,
        '\uff11': 1,
        '\uff12': 2,
        '\uff13': 3,
        '\uff14': 4,
        '\uff15': 5,
        '\uff16': 6,
        '\uff17': 7,
        '\uff18': 8,
        '\uff19': 9,
        '\uff21': 10,
        '\uff22': 11,
        '\uff23': 12,
        '\uff24': 13,
        '\uff25': 14,
        '\uff26': 15,
        '\uff41': 10,
        '\uff42': 11,
        '\uff43': 12,
        '\uff44': 13,
        '\uff45': 14,
        '\uff46': 15,
        }
