
from .utils import new_topol as topol, find_loops
from .grammar import NTSYM, TSYM, OPT, ANY, SEQ, ARG, NAMED, TERMINAL, EOF, Ref
from .generator import *

EPSILON = object()
MARK = object()

class LRProduction:
    def __init__(self, lhs, rhs, args, kwargs, action, kwds, src):
        self.lhs = lhs
        self.rhs = rhs
        self.action = action
        kwargs = kwargs.copy()
        kwargs.update(kwds.pop('kwargs',{}))
        args = args + kwds.pop('args',[])
        self.args = (args, kwargs)
        self.kwds = kwds
        self.src = src

class LR0Item:
    def __init__(self, p):
        self.p = p
        self.i = 0

    def get(self):
        if self.i >= len(self.p.rhs): return None
        return self.p.rhs[self.i]

    def next(self):
        c = LR0Item(self.p)
        c.i = self.i+1
        return c

    def lr1(self, f):
        c = LR1Item(self.p, f)
        c.i = self.i
        return c

    def terminal(self):
        return self.i >= len(self.p.rhs)

    def final(self):
        if not self.p.rhs: return False
        return isinstance(self.p.rhs[-1], TSYM) and self.p.rhs[-1].v is EOF

    def source(self):
        if not self.p.rhs:
            return None
        if self.i == len(self.p.rhs):
            t = self.p.rhs[-1].src
            if t is None: return None
            return (t.s, t.e)
        else:
            t = self.p.rhs[self.i].src
            if t is None: return None
            return (t.s, t.b)

    @staticmethod
    def closure_items(p, i, gen):
        return (LR0Item(p),)

    def __repr__(self):
        return "[{} -> {} . {}]".format(
            self.p.lhs,
            ' '.join(repr(x) for x in self.p.rhs[:self.i]),
            ' '.join(repr(x) for x in self.p.rhs[self.i:]))

    def __eq__(self, other):
        return self.p == other.p and self.i == other.i

    def __hash__(self):
        return hash((self.p, self.i))

#    def copy(self):
#        c = LR0Item(self.p)
#        c.i = self.i
#        return c

class LR1Item (LR0Item):
    def __init__(self, p, f=None):
        super().__init__(p)
        self.f = f

    def next(self):
        c = LR1Item(self.p, f=self.f)
        c.i = self.i+1
        return c

    def lr0(self):
        c = LR0Item(self.p)
        c.i = self.i
        return c

    @staticmethod
    def closure_items(p, i, gen):
        for t in gen.first(i.lr0().next()):
            if t is EPSILON:
                t = i.f
            elif isinstance(t, NTSYM):
                continue
            yield LR1Item(p, f=t)

    def __repr__(self):
        return super().__repr__()[:-1]+" + {}]".format(self.f)

    def __eq__(self, other):
        return self.p == other.p and self.i == other.i and \
            (self.f == other.f or (self.f is None and other.f is None))

    def __hash__(self):
        return hash((self.p, self.i, self.f))

class LRPreparation (Preparation):
    def run(self):
        if isinstance(self.grammar.START, tuple):
            self.start = self.grammar.START
        else:
            self.start = (self.grammar.START,)
        self.seq_cache = []
        for p in self.grammar.PRODUCTIONS:
            self.do_rule(p.lhs, p.rhs, p.action, p.kwds, p.src)

    def do_rule(self, lhs, rhs, act, kwds, src=None):
        stack = [([rhs], [], {})]
        while stack:
            ss, pr, nm = stack[-1]
            if ss:
                ss.pop().process(self, stack)
            else:
                stack.pop()
                pargs = nm.pop('-',[])+[]
                if act is None:
                    uact = self.default_action(pr, pargs, nm)
                else:
                    uact = act
                self.add_prod(lhs, pr, pargs, nm, uact, kwds, src)

    def default_action(self, rhs, pargs, kargs):
        if kargs:
            if pargs:
                raise GrammarError("Cannot create a default action with both "
                                   "positional and named arguments")
            return 'DICT'
        elif len(pargs) == 1:
            return 'COPY'
        elif pargs:
            return 'TUPLE'
        else:
            n = len(rhs)
            pargs[:] = [Ref(x) for x in range(n)]
            if n == 1:
                return 'COPY'
            return 'TUPLE'

    def add_prod(self, lhs, rhs, args, kwargs, action, kwds, src=None):
        p = LRProduction(lhs, rhs, args, kwargs, action, kwds, src)
        p.n = len(self.rules)
        self.rules.append(p)
        lhs.rules.append(p)
        return p

    def copy_nm(self, nm):
        nm = nm.copy()
        if '-' in nm: nm['-'] = nm['-'] + []
        if '!!' in nm: nm['!!'] = nm['!!'] + []
        return nm

    def process_terminal(self, obj, stack):
        stack[-1][1].append(obj)

    def process_ntsym(self, obj, stack):
        self.ntsyms.add(obj.v)
        stack[-1][1].append(obj)

    def process_arg(self, obj, stack):
        if isinstance(obj.v, NTSYM) or isinstance(obj.v, TERMINAL):
            obj.v.process(self, stack)
            stack[-1][2].setdefault('-', []).append(Ref(len(stack[-1][1])-1))
        elif isinstance(obj.v, SEQ):
            self.process_seq(obj.v, stack, True)
            stack[-1][2].setdefault('-', []).append(Ref(len(stack[-1][1])-1))
        elif isinstance(obj.v, OPT):
            ss, pr, nm = stack[-1]
            nm2 = self.copy_nm(nm)
            nm2.setdefault('-', []).append(True)
            nm.setdefault('-', []).append(False)
            stack.append((ss+[obj.v.v], pr+[], nm2))
        else:
            raise GrammarError("That object can't be an argument", obj.src)

    def process_dictarg(self, obj, stack):
        if not isinstance(obj.v, NTSYM):
            raise GrammarError("That object can't be a dict argument", obj.src)
        obj.v.process(self, stack)
        stack[-1][2].setdefault('!!', []).append(Ref(len(stack[-1][1])-1))

    def process_named(self, obj, stack):
        if isinstance(obj.v, NTSYM) or isinstance(obj.v, TERMINAL):
            obj.v.process(self, stack)
            stack[-1][2][obj.n] = Ref(len(stack[-1][1])-1)
        elif isinstance(obj.v, SEQ):
            self.process_seq(obj.v, stack, arg=obj.n)
            stack[-1][2][obj.n] = Ref(len(stack[-1][1])-1)
        elif isinstance(obj.v, OPT):
            ss, pr, nm = stack[-1]
            nm2 = self.copy_nm(nm)
            nm2[obj.n] = True
            nm[obj.n] = False
            stack.append((ss+[obj.v.v], pr+[], nm2))
        else:
            raise GrammarError("Can't name that object", obj.n, obj.src)

    def process_cat(self, obj, stack):
        stack[-1][0].append(obj.b)
        stack[-1][0].append(obj.a)

    def process_opt(self, obj, stack):
        ss, pr, nm = stack[-1]
        nm2 = self.copy_nm(nm)
        if isinstance(obj.v, NAMED):
            nm[obj.v.n] = None
        stack.append((ss+[obj.v], pr+[], nm2))

    def process_alt(self, obj, stack):
        ss, pr, nm = stack[-1]
        stack.append((ss+[obj.a], pr+[], self.copy_nm(nm)))
        stack[-2][0].append(obj.b)

    def process_seq(self, obj, stack, arg=None):
        if not hasattr(obj, 'nt'):
            obj.nt = None
            for s in self.seq_cache:
                if s.v.deep_eq(obj.v) and s.sep == obj.sep:
                    obj.nt = s.nt
            if obj.nt is None:
                obj.nt = self.create_seq(obj)
                self.seq_cache.append(obj)
        if obj.min == 0:
            ss, pr, nm = stack[-1]
            nm2 = self.copy_nm(nm)
            if arg is True:
                nm.setdefault('-', []).append([])
            elif arg:
                nm[arg] = []
            stack.append((ss+[], pr+[obj.nt], nm2))
        else:
            stack[-1][1].append(obj.nt)

    def create_seq(self, obj):
        USE_MARK = False
        seq = self.new_nt()
        if isinstance(obj.v, ARG) and isinstance(obj.v.v, (TERMINAL, NTSYM)):
            bo = obj.v.v
        else:
            body = self.new_nt()
            self.do_rule(body, obj.v, None, {})
            bo = NTSYM(body, src=obj.v.src)
        if USE_MARK:
            mark = self.new_nt()
            self.add_prod(mark, [], [], {}, "TUPLE", {})
            self.add_prod(seq, [NTSYM(mark, src=obj.src), bo],
                          [Ref(1)], {}, "LIST", {})
        else:
            self.add_prod(seq, [bo], [Ref(0)], {}, "LIST", {})
        if obj.sep:
            self.add_prod(seq, [NTSYM(seq, src=obj.src), TEXT(obj.sep), bo],
                          [Ref(0),Ref(2)], {}, "APPEND", {})
        else:
            self.add_prod(seq, [NTSYM(seq, src=obj.src), bo],
                          [Ref(0),Ref(1)], {}, "APPEND", {})
        return NTSYM(seq, src=obj.src)

    def debug(self):
        for r in self.rules:
            print(r.lhs, '->', r.rhs)
        print()

class conflict_resolver:
    def __init__(self, fn):
        self.fn = fn

    def __call__(self, *args):
        return self.fn(*args)

class StdConflictResolvers:
    @conflict_resolver
    def conflict_precedence(tbl, s, t, si, ri):
        if not si: return False
        sp = si[0].p.kwds.get('precedence')
        if sp is None or any(x.p.kwds.get('precedence') is not sp for x in si):
            return False
        c,a = tbl.states[s].action[t]
        r = a[0]
        cp = sp
        for p,j in zip(ri,a[1:]):
            rp = p.kwds.get('precedence')
            if rp is None: return False
            if rp == cp:
                if rp.LEFT:
                    r = j
                    cp = rp
            elif rp.n < cp.n:
                r = j
                cp = rp
        tbl.states[s].action[t] = r
        return True

    @conflict_resolver
    def conflict_specified_shift(tbl, s, t, si, ri):
        if not si: return False
        for p in ri:
            rs = p.kwds.get('resolve_shift')
            if not rs or t != rs: return False
        tbl.states[s].action[t] = tbl.states[s].action[t][1][0]
        return True

    @conflict_resolver
    def conflict_specified_resolve(tbl, s, t, si, ri):
        if si: return False
        pref = None
        for p,a in zip(ri, tbl.states[s].action[t][1]):
            rs = p.kwds.get('resolve_prefer')
            if rs: pref = a
        if not pref: return False
        tbl.states[s].action[t] = pref
        return True

    def default_resolver(tbl, s, t, si, ri):
        print("Conflict on {}:".format(t))
        print("   for state",tbl.states[s].sid)
        for p in tbl.states[s].src:
            print("   ", p)
        for p in si:
            print("  s", p)
        for p in ri:
            print("  r", p.lhs, "->", p.rhs)
        print()
        return False

class GLRConflictResolvers (StdConflictResolvers):
    def default_resolver(tbl, s, t, si, ri):
        return True

class ShiftDefaultResolver:
    def default_resolver(tbl, s, t, si, ri):
        if si:
            tbl.states[s].action[t] = tbl.states[s].action[t][1][0]
            return True
        return False

class ContextFreeParserGenerator (ParserGenerator):
    PREPARATION = LRPreparation

    def start_state(self, ss):
        snt = self.pgr.new_nt()
        from .src_string import StringNodeSourceTrack
        s = "START -> {} <EOF>".format(ss)
        sl = len(s)
        s1 = StringNodeSourceTrack(s, 9, sl-6)
        s2 = StringNodeSourceTrack(s, sl-5, sl)
        p = self.pgr.add_prod(snt, [NTSYM(ss, src=s1),
                                    TSYM(EOF, src=s2)],
                              [], {}, None, {})
        return self.item(p)

    def transitions(self, st):
        o = st.get()
        if o is not None:
            yield (o, st.next())

    def closure(self, ss):
        c = set(ss)
        todo = list(ss)
        while todo:
            s = todo.pop()
            o = s.get()
            if o is None or isinstance(o, TERMINAL): continue
            for p in o.v.rules:
                for i in s.closure_items(p, s, self):
                    if i not in c:
                        c.add(i)
                        todo.append(i)
        return c

    FIRST_CACHE = {}

    def first(self, i):
        f = self.FIRST_CACHE.get((self, i))
        if f is not None: return f
        self.FIRST_CACHE[(self, i)] = set()
        f = { EPSILON }
        j = i
        while not j.terminal() and EPSILON in f:
            f.remove(EPSILON)
            t = j.get()
            if not isinstance(t, NTSYM):
                f.add(t)
            else:
                for p in t.v.rules:
                    f.update(self.first(LR0Item(p)))
            j = j.next()
        self.FIRST_CACHE[(self, i)] = f
        return f

    def display_raw_table(self, start, tbl):
        for s in tbl:
            print("STATE {} {}::".format(self.sid[s],
                                         ('START ' if s in start else '')))
            for i in s:
                t = i.get()
                print("  i ({},{}) {}".format(i.p.n,i.i,i))
                if t is None:
                    print("    ",i.p.rhs[-1].src.s,'.')
                else:
                    print("    ",t.src)
            for t in tbl[s]:
                d = tbl[s][t]
                print("  t ",t," -> ",(self.sid[d] if d is not None else '---'))
            print()

class ShiftReduceParserGenerator (ContextFreeParserGenerator):
    from .eng_sr import ShiftReduceTable as TABLE_CLASS, \
        ShiftReduceEngine as ENGINE
    CONFLICT_RESOLVERS = StdConflictResolvers

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.resolvers = []
        for b in self.pbases+(self.CONFLICT_RESOLVERS,):
            if hasattr(b, 'default_resolver'):
                self.default_resolver = b.default_resolver
                break
        for b in self.pbases+(StdConflictResolvers,):
            for v in b.__dict__.values():
                if isinstance(v, conflict_resolver):
                    self.resolvers.append(v)

    def table(self):
        rss, rawtbl = super().table()
        self.sid = { s:n for n,s in enumerate(rawtbl) }
        if hasattr(self.grammar, 'DEBUG'):
            self.display_raw_table(rss, rawtbl)
        self.prepare_reduce_symbols(rss, rawtbl)
        tbl = self.TABLE_CLASS(self.grammar.__name__)
        for rs in rss:
            tbl.ensure_state(rs, self)
        for rs,trt in rawtbl.items():
            self.table_row(rs, trt, tbl)
        if tbl.conflicts:
            self.resolve_all_conflicts(tbl)
        return tbl

    def table_row(self, rs, trt, tbl):
        st = tbl.ensure_state(rs, self)
        tbl.states[st].sid = self.sid[rs]
        for sm,tr in trt.items():
            if not tr: continue
            ds = tbl.ensure_state(tr, self)
            if isinstance(sm, NTSYM):
                tbl.add_goto(st, sm, ds)
            elif isinstance(sm, TSYM) and sm.v is EOF:
                tbl.add_accept(st, sm)
            else:
                tbl.add_shift(st, sm, ds)
        for it in self.state_items(rs):
            if not it.terminal() or it.final(): continue
            p = tbl.ensure_rule(it.p)
            for sm in self.reduce_symbols(tbl.states[st].raw, it):
                tbl.add_reduce(st, sm, p)

    def resolve_all_conflicts(self, tbl):
        r = True
        for s,t in tbl.conflicts:
            st = tbl.states[s]
            cc,aa = st.action[t]
            if aa[0][0] == 's':
                si = [i for i in self.state_items(st.raw)
                      if i.get() and i.get().matches(t, self.grammar)]
                ri = [tbl.rules[p[1]].raw for p in aa[1:]]
            else:
                si = []
                ri = [tbl.rules[p[1]].raw for p in aa]
            r = self.resolve_conflict(tbl, s, t, si, ri) and r
        if not r:
            raise Exception("Unresolved Conflicts")

    def resolve_conflict(self, tbl, s, t, si, ri):
        for cr in self.resolvers:
            r = cr(tbl, s, t, si, ri)
            if r: return True
        return self.default_resolver(tbl, s, t, si, ri)

class LL1ParserGenerator (ShiftReduceParserGenerator):
    item = LR0Item

class LR0ParserGenerator (SetsMethod, ShiftReduceParserGenerator):
    item = LR0Item

    def prepare_reduce_symbols(self, start, tbl):
        pass

    def reduce_symbols(self, st, it):
        return (ANY(),)

class LR1ParserGenerator (SetsMethod, ShiftReduceParserGenerator):
    item = LR1Item

    def prepare_reduce_symbols(self, start, tbl):
        pass

    def reduce_symbols(self, st, it):
        return (it.f,)

class LALRParserGenerator (SetsMethod, ShiftReduceParserGenerator):
    item = LR0Item

    def prepare_reduce_symbols(self, start, tbl):
        c = { (s, tuple(s)[0]) : ( { EOF }, set() ) for s in start }
        for s in tbl:
            for i in s:
                j = i.lr1(MARK)
                for jj in self.state_items({j}):
                    if len(jj.p.rhs) == 0:
                        sss = { s }
                        ii = jj.lr0()
                    elif jj.terminal():
                        continue
                    else:
                        t = jj.get()
                        if not isinstance(t, SET):
                            sss = { tbl[s][t] }
                        elif t.inv:
                            sss = { tbl[s][ANY()] }
                        else:
                            sss = { tbl[s][TEXT(tt)] for tt in t.s }
                        ii = jj.lr0().next()
                    for ss in sss:
                        c.setdefault((ss,ii), (set(), set()))
                        if jj.f is not MARK:
                            c[ss,ii][0].add(jj.f)
                        elif s != ss and i != ii:
                            c[ss,ii][1].add((s,i))
        if hasattr(self.grammar, 'DEBUG'):
            for s,i in c:
                d,id = c[s,i]
                si = ('({},{},{})'.format(self.sid[xs],xi.p.n,xi.i) for xs, xi in id)
                print('({},{},{}) : {} , {}'.format(self.sid[s],i.p.n,i.i, d,','.join(si)))
            print('\n***********\n\n')
            def report(lp):
                s,i = lp[0]
                lps = ('({},{},{})'.format(self.sid[xs],xi.p.n,xi.i) for xs, xi in lp[2])
                print('({},{},{}) : {}'.format(self.sid[s],i.p.n,i.i,','.join(lps)))
            find_loops(c, report)
            print()
            f = open("lookahead.dot", "w")
            f.write("digraph Lookahead {\n")
            for s,i in c:
                d,id = c[s,i]
                for xs,xi in id:
                    f.write('  "{},{},{}" -> "{},{},{}";\n'.format(
                            self.sid[s],i.p.n,i.i, self.sid[xs],xi.p.n,xi.i))
            f.write("}\n")
            f.close()
        self.lookahead = topol(c)
        """
        for s,i in self.lookahead:
            print('({},{},{}) : {}'.format(self.sid[s],i.p.n,i.i,self.lookahead[s,i]))
        print('\n***********\n\n')
        """

    def reduce_symbols(self, st, it):
        return self.lookahead[st,it]

class GLRParserGenerator (LALRParserGenerator):
    from .eng_glr import GLREngine as ENGINE
    CONFLICT_RESOLVERS = GLRConflictResolvers
