
from .utils import topol, frozendict
from .generator import *

class RegularProduction:
    def __init__(self, lhs, t, nt=None, action=None, src=None):
        self.lhs = lhs
        self.t = t
        self.nt = nt
        self.action = action
        self.src = src

    def __repr__(self):
        return '[{} -> {} {}]'.format(self.lhs, self.t, self.nt)

class RegularPreparation (Preparation):
    def run(self):
        self.start = []
        self.end = self.new_nt()
        stack = {}
        if isinstance(self.grammar.START, tuple):
            ss = self.grammar.START
        else:
            ss = (self.grammar.START,)
        for s in ss:
            start = self.new_nt()
            self.start.append(start)
            stack[s] = start
            self.do_symbol(s, start, self.end, stack)

    def do_symbol(self, gsym, start, end, stack):
        for p in self.grammar.PRODUCTIONS:
            if p.lhs is not gsym: continue
            p.rhs.process(self, start, end, stack, True, p)

    def new_prod(self, lhs, t, nt, action=None, src=None):
        p = RegularProduction(lhs, t, nt, action, src)
        self.rules.append(p)
        lhs.rules.append(p)

    def new_epsilon(self, lhs, rhs, action=None, src=None):
        p = RegularProduction(lhs, None, rhs, action, src)
        self.rules.append(p)
        lhs.rules.append(p)

    def process_terminal(self, obj, start, end, stack, final, action=None):
        self.new_prod(start, obj, end, action, src=obj.src)

    def process_ntsym(self, obj, start, end, stack, final, action=None):
        if obj.v in stack:
            if not final:
                raise GeneratorError("Not a Regular Grammar", obj.v)
            self.new_epsilon(start, stack[obj.v], src=obj.src)
        else:
            ns = stack.copy()
            ns[obj.v] = start
            self.do_symbol(obj.v, start, end, ns) # what about our action!

    def process_arg(self, obj, start, end, stack, final, action=None):
        ### we are ignoring args for now!
        obj.v.process(self, start, end, stack, final, action)

    def process_dictarg(self, obj, start, end, stack, final, action=None):
        ### we are ignoring args for now!
        obj.v.process(self, start, end, stack, final, action)

    def process_named(self, obj, start, end, stack, final, action=None):
        ### we are ignoring the name for now!
        obj.v.process(self, start, end, stack, final, action)

    def process_cat(self, obj, start, end, stack, final, action=None):
        mid = self.new_nt()
        obj.a.process(self, start, mid, stack, False)
        obj.b.process(self, mid, end, stack, final, action)

    def process_opt(self, obj, start, end, stack, final, action=None):
        obj.v.process(self, start, end, stack, final, action)
        self.new_epsilon(start, end, action, src=obj.src)

    def process_alt(self, obj, start, end, stack, final, action=None):
        obj.a.process(self, start, end, stack, final, action)
        obj.b.process(self, start, end, stack, final, action)

    def process_seq(self, obj, start, end, stack, final, action=None):
        a = self.new_nt()
        b = self.new_nt()
        self.new_epsilon(start, a, src=obj.src)
        obj.v.process(self, a, b, stack, False)
        self.new_epsilon(b, end, action, src=obj.src)
        if obj.min == 0:
            self.new_epsilon(start, end, action, src=obj.src)
        if obj.sep is not None:
            obj.sep.process(self, b, a, stack)
        else:
            self.new_epsilon(b, a, src=obj.src)

class DRegularParserGenerator (ParserGenerator):
    PREPARATION = RegularPreparation
    from .eng_fsm import FSMTable as TABLE_CLASS, FiniteStateMachine as ENGINE

    def start_state(self, ss):
        return ss

    def transitions(self, st):
        return ((p.t, p.nt) for p in st.rules if p.t is not None)

    def closure(self, ss):
        c = set(ss)
        todo = list(ss)
        while todo:
            s = todo.pop()
            for p in s.rules:
                if p.t is not None or p.nt is None or p.nt in c: continue
                c.add(p.nt)
                todo.append(p.nt)
        return c

    def conflict(self, st, t, x, ns):
        print("Conflict", st, t)
        raise SystemExit

    def table(self):
        rss, rawtbl = super().table()
#        print()
#        for s,ts in rawtbl.items():
#            print("STATE {}".format(', '.join(str(x) for x in s)))
#            for t,ds in ts.items():
#                if ds is not None:
#                    print("  {} : {}".format(t, ', '.join(str(x) for x in ds)))
#                else:
#                    print("  {} : error".format(t))
        tbl = self.TABLE_CLASS(self.grammar.__name__, len(rss))
        for rs in rss:
            tbl.ensure_state(rs, self)
        for rs,trt in rawtbl.items():
            self.table_row(rs, trt, tbl)
        return tbl

    def table_row(self, rs, trt, tbl):
        st = tbl.ensure_state(rs, self)
        for sm,tr in trt.items():
            if not tr:
                tbl.add(st, sm, 'e', None)
                continue
            ds = tbl.ensure_state(tr, self)
            actions = set()
            for r in self.state_items(rs):
                for p in r.rules:
                    if p.t and p.t.matches(sm.match(), self.grammar):
                        if p.action is not None:
                            actions.add(p.action)
                        actions.update({x.action for x in rs[r]})
            actions.update({x.action for x in tr.get(self.pgr.end, ())})
            if len(actions) == 0:
                action = None
            elif len(actions) == 1:
                action = actions.pop().action
            else:
                action = self.action_conflict(rs, sm, actions)
            tbl.add(st, sm, ds, action)

    def action_conflict(self, st, sm, actions):
        for b in self.pbases:
            if hasattr(b, 'action_conflict'):
                return b.action_conflict(st, sm, actions)
        if sm.v == '.':
            print('---')
        print("Conflict on {!r}::".format(sm))
        for a in actions:
            print("    ",a.src.s)
        print("  in state::")
        for r in self.state_items(st):
            for p in r.rules:
                print("    ",p.src.s[:p.src.b],'.',p.src.s[p.src.b:])
        print()

class RegularParserGenerator (SetsMethod, DRegularParserGenerator):
    def old_make_state(self, ss):
        st = []
        for s in super().state_items(ss):
            if any(p.t is not None for p in s.rules) or not s.rules:
                st.append(s)
        return frozenset(st)

    def state_items(self, st):
        return st

    def make_state(self, ss):
        c = { s:(set(),set()) for s in ss }
        todo = list(ss)
        while todo:
            s = todo.pop()
            for p in s.rules:
                if p.t is not None or p.nt is None: continue
                if p.nt not in c:
                    c[p.nt] = (set(),set())
                    todo.append(p.nt)
                if p.action:
                    c[p.nt][0].add(p)
                c[p.nt][1].add(s)
        return frozendict((s,frozenset(a)) for s,a in topol(c).items()
                          if any(p.t is not None for p in s.rules) or
                          not s.rules)

class ConflictFirstListed:
    
    @staticmethod
    def action_conflict(st, sm, actions):
        act = list(actions)
        act.sort(key=lambda x:x.n)
        return act[0].action
