
### Symbols

ANON = '<*anonymous*>'

class Symbol:
    def __init__(self, name=ANON):
        self.name = name
        self.anon = False

    def __repr__(self):
        return self.name

#    def match(self):
#        return self

    def etree(self, p, f, attrs):
        return f(p, "symbol", attrs, value=self.name)

class NonTerminal (Symbol):
    "A non-terminal symbol"

    def __init__(self, name=ANON):
        self.name = name
        self.rules = []
        self.anon = False

    def source(self):
        return self.name

    def __lt__(self, other):
        return id(self) < id(other)

EOF = Symbol(name='EOF')

### Other Grammar Objects

class Category:
    pass

class RMacro:
    def __init__(self, src):
        self.src = src
        self.o = {}

    def compile(self, grammar):
        o = self.o.get(grammar)
        if o is None:
            o = rparse(self.src, grammar, st=1)
            self.o[grammar] = o
        return o

### GProduction class

class GProduction:
    def __init__(self, lhs=None, rhs=None, kwds={}, srcs=None, parse=None):
        self.action = kwds.pop('action', None)
        self.kwds = kwds
        self.lhs = lhs
        self.rhs = rhs
        self.srcs = srcs
        self.parse = parse
        self.src = None
        self.decor = False

    def __call__(self, actfn):
        "GProductions can be used as decorators of the action function"
        self.action = actfn
        self.decor = True
        return self

    def ensure(self, gcls):
        if self.lhs is None:
            self.lhs, self.rhs, self.src = self.parse(self.srcs, gcls)

### Precedence

class Precedence:
    COUNT = 0

    def __init__(self, cls=0):
        self.n = cls*100000 + Precedence.COUNT
        Precedence.COUNT += 1

class LeftAssociative (Precedence):
    "A left-associatve precedence object for expression grammars"
    LEFT = True
 
class RightAssociative (Precedence):
    "A right-associative precedence object for expression grammars"
    LEFT = False

### Rule objects

class TERMINAL:
    def process(self, callo, *args):
        callo.process_terminal(self, *args)

    def deep_eq(self, other):
        return self == other

class ANY (TERMINAL):
    def __init__(self, src=None):
        self.src = src

    def __eq__(self, other):
        return isinstance(other, ANY)

    def __hash__(self):
        return hash(ANY)

    def __repr__(self):
        return 'ANY()'

    def match(self):
        return self

    def matches(self, t, gr):
        return True

    def etree(self, p, f, attrs):
        return f(p, "any", attrs)

class TYP (TERMINAL):
    def __init__(self, c, src=None):
        self.c = c
        self.src = src

    def __eq__(self, other):
        return isinstance(other, TYP) and self.c == other.c

    def __hash__(self):
        return hash(self.c)

    def __repr__(self):
        return 'TYP({!r})'.format(self.c)

    def match(self):
        return self

    def matches(self, t, gr):
        if isinstance(t, TYP):
            return t.c == self.c
        return gr.category(t) == self.c

    def etree(self, p, f, attrs):
        return f(p, "type", attrs, value=self.c)

class SET (TERMINAL):
    def __init__(self, s, inv=False, src=None):
        self.s = frozenset(s)
        self.inv = inv
        self.src = src

    def __eq__(self, other):
        return isinstance(other, SET) and self.s == other.s and self.inv == other.inv

    def __hash__(self):
        return hash((self.s, self.inv))

    def __repr__(self):
        return '[{}{}]'.format(('^' if self.inv else ''), ''.join(self.s))

    def matches(self, t, gr):
        if isinstance(t, TYP): return self.inv
        if self.inv: return t not in self.s
        return t in self.s

class TEXT (TERMINAL):
    def __init__(self, v, src=None):
        self.v = v
        self.src = src

    def __eq__(self, other):
        return isinstance(other, TEXT) and self.v == other.v

    def __hash__(self):
        return hash(self.v)

    def __repr__(self):
        return repr(self.v)

    def match(self):
        return self.v

    def matches(self, t, gr):
        return self.v == t

class TSYM (TERMINAL):
    def __init__(self, v, src=None):
        self.v = v
        self.src = src

    def __eq__(self, other):
        return isinstance(other, TSYM) and self.v == other.v

    def __hash__(self):
        return hash(self.v)

    def __repr__(self):
        return repr(self.v)

    def match(self):
        return self.v

    def matches(self, t, gr):
        return self.v == t

class NTSYM:
    def __init__(self, v, src=None):
        self.v = v
        self.src = src

    def process(self, callo, *args):
        callo.process_ntsym(self, *args)

    def __eq__(self, other):
        return isinstance(other, NTSYM) and self.v == other.v

    def __hash__(self):
        return hash(self.v)

    def __repr__(self):
        return repr(self.v)

    def match(self):
        return self.v

    def matches(self, t, gr):
        return False

    def deep_eq(self, other):
        return isinstance(other, NTSYM) and self.v == other.v

class ARG:
    def __init__(self, v, src=None):
        self.v = v
        self.src = src

    def process(self, callo, *args):
        callo.process_arg(self, *args)

    def deep_eq(self, other):
        return isinstance(other, ARG) and self.v.deep_eq(other.v)

class DICTARG:
    def __init__(self, v, src=None):
        self.v = v
        self.src = src

    def process(self, callo, *args):
        callo.process_dictarg(self, *args)

    def deep_eq(self, other):
        return isinstance(other, DICTARG) and self.v.deep_eq(other.v)

class NAMED:
    def __init__(self, v, n, src=None):
        self.v = v
        self.n = n
        self.src = src

    def process(self, callo, *args):
        callo.process_named(self, *args)

    def deep_eq(self, other):
        return isinstance(other, NAMED) and self.v.deep_eq(other.v) \
            and self.n == other.n

class CAT:
    def __init__(self, a, b, src=None):
        self.a = a
        self.b = b
        self.src = src

    def process(self, callo, *args):
        callo.process_cat(self, *args)

    def deep_eq(self, other):
        return isinstance(other, CAT) and self.a.deep_eq(other.a) \
            and self.b.deep_eq(other.b)

class OPT:
    def __init__(self, v, src=None):
        self.v = v
        self.src = src

    def process(self, callo, *args):
        callo.process_opt(self, *args)

    def deep_eq(self, other):
        return isinstance(other, OPT) and self.v.deep_eq(other.v)

class ALT:
    def __init__(self, a, b, src=None):
        self.a = a
        self.b = b
        self.src = src

    def process(self, callo, *args):
        callo.process_alt(self, *args)

    def deep_eq(self, other):
        return isinstance(other, ALT) and self.a.deep_eq(other.a) \
            and self.b.deep_eq(other.b)

class SEQ:
    def __init__(self, v, min=0, sep=None, src=None):
        self.v = v
        self.min = min
        self.sep = sep
        self.src = src

    def process(self, callo, *args):
        callo.process_seq(self, *args)

    def deep_eq(self, other):
        return isinstance(other, SEQ) and self.v.deep_eq(other.v) \
            and self.min == other.min and self.sep == other.sep

### Production generators

PATTERN_PARSE = []

def ensure_tables():
    if PATTERN_PARSE: return
    from . import ptn_tables

def pparse(srcs, gcls):
    from .src_string import StringTraceSource
    EBNFLexical, EBNFParser = PATTERN_PARSE[1:3]
    parser = EBNFParser(EBNFLexical(StringTraceSource(srcs)), gcls)
    return parser.run()

def rparse(srcs, gcls, st=0):
    from .src_string import StringTraceSource
    RegexParser = PATTERN_PARSE[0]
    parser = RegexParser(StringTraceSource(srcs), gcls)
    return parser.run(st)

def p(rule, **kwds):
    "create a grammar rule using EBNF syntax"
    return GProduction(srcs=rule, parse=pparse, kwds=kwds)

def r(rule, **kwds):
    "create a grammar rule using Regex syntax"
    return GProduction(srcs=rule, parse=rparse, kwds=kwds)

### Internal

class Ref (int):
    def __repr__(self):
        return 'Ref({})'.format(super().__repr__())

### Grammar class

class Attrs (dict):
    def __init__(self):
        self.prods = []

    def __setitem__(self, k, v):
        if isinstance(v, GProduction):
            self.prods.append(v)
            if v.decor:
                f = v.action
                v.action = k
                v = f
        elif isinstance(v, Symbol) and v.name is ANON:
            v.name = k
        super().__setitem__(k, v)

class GType (type):
    @staticmethod
    def __prepare__(name, bases, *kwds):
        return Attrs()

    def __new__(mcls, name, bases, attrs):
        cls = super().__new__(mcls, name, bases, attrs)
        if not isinstance(attrs, Attrs): return cls
        cls.PRODUCTIONS = attrs.prods
        for b in bases:
            if hasattr(b, 'PRODUCTIONS'):
                cls.PRODUCTIONS.extend(b.PRODUCTIONS)
        return cls

class Grammar (metaclass=GType):
    """The Grammar class is used to describe grammars.

    Both Regular and Context-free grammars are supported.
    The Grammar should contain production rules and symbol
    declarations.
    """

def grammar_factory(name, parents, *prods, **attrs):
    a = Attrs()
    a.update(attrs)
    a.prods.extend(list(prods))
    return type(name, tuple(parents), a)
