
from functools import reduce

from grammar import RegularParserGenerator, LALRParserGenerator, \
    LexicalAnalyser, Token, MatchToken, SourceTrace, PythonWriter
from grammar.grammar import *

## Boot Regular Expression parser.  Manually parsed

class RegexBootGrammar (Grammar):
    production = NonTerminal()
    regex = NonTerminal()
    symbol = NonTerminal()
    srange = NonTerminal()
    relem = NonTerminal()

    p_value = LeftAssociative()
    p_unary = LeftAssociative()
    p_cat = LeftAssociative()
    p_alt = LeftAssociative()

    START = production
    ESCAPE = { 'n': '\n', 'r': '\n', 't': '\t' }

    @GProduction(production, CAT(ARG(NTSYM(symbol)),
                                 CAT(SEQ(TEXT(' ')),
                                     CAT(TEXT('-'),
                                         CAT(TEXT('>'),
                                             CAT(SEQ(TEXT(' ')),
                                                 ARG(NTSYM(regex))))))))
    def match_production(self, ls, rhs, src):
        lhs = self.get_symbol(ls)
        if not isinstance(lhs, NonTerminal):
            raise GrammarError("Left hand side must be a non-terminal", ls)
        return (lhs, rhs, src)

    @GProduction(regex, CAT(ARG(NTSYM(regex)),
                            CAT(TEXT('|'), ARG(NTSYM(regex)))),
                 {'precedence':p_alt})
    def match_alt(self, a, b, src):
        return ALT(a, b, src=src)

    @GProduction(regex, CAT(ARG(NTSYM(regex)), ARG(NTSYM(regex))),
                 {'precedence':p_cat})
    def match_cat(self, a, b, src):
        return CAT(a, b, src=src)

    @GProduction(regex, CAT(ARG(NTSYM(regex)), TEXT('*')),
                 {'precedence':p_unary})
    def match_seq0(self, v, src):
        return SEQ(v, min=0, src=src)

    @GProduction(regex, CAT(ARG(NTSYM(regex)), TEXT('+')),
                 {'precedence':p_unary})
    def match_seq1(self, v, src):
        return SEQ(v, min=1, src=src)

    @GProduction(regex, CAT(ARG(NTSYM(regex)), TEXT('?')),
                 {'precedence':p_unary})
    def match_opt(self, v, src):
        return OPT(v, src=src)

    @GProduction(regex, CAT(ARG(NTSYM(regex)), TEXT('!')),
                 {'precedence':p_unary})
    def match_arg(self, v, src):
        return ARG(v, src=src)

    @GProduction(regex, CAT(ARG(NTSYM(regex)),
                            CAT(TEXT(':'), ARG(NTSYM(symbol)))),
                 {'precedence':p_unary})
    def match_named(self, v, n, src):
        return NAMED(v, n, src=src)

    _ = GProduction(regex, CAT(TEXT('('),
                               CAT(ARG(NTSYM(regex)), TEXT(')'))),
                    {'precedence':p_value, 'action':'COPY'})

    @GProduction(regex, CAT(TEXT('['),
                            CAT(NAMED(OPT(TEXT('^')),'inv'),
                                CAT(ARG(SEQ(ARG(NTSYM(srange)), min=1)),
                                    TEXT(']')))), {'precedence':p_value})
    def match_set(self, s, inv, src):
        return SET(reduce(set.union, s), inv, src=src)

    @GProduction(regex, CAT(TEXT('{'),
                            CAT(ARG(NTSYM(symbol)),
                                CAT(OPT(CAT(TEXT(':'),
                                            ARG(NTSYM(symbol)))),
                                    TEXT('}')))), {'precedence':p_value})
    def match_symbol(self, c, n=None, src=None):
        v = self.get_symbol(c)
        if isinstance(v, NonTerminal):
            o = NTSYM(v, src=src)
        elif isinstance(v, Symbol):
            o = TSYM(v, src=src)
        else:
            o = TYP(c, src=src)
        if n is None: return o
        return NAMED(o, n, src=src)

    @GProduction(regex, CAT(TEXT('\\'), ARG(ANY())), {'precedence':p_value})
    def match_escape(self, v, src):
        return TEXT(self.ESCAPE.get(v, v), src=src)

    @GProduction(regex, TEXT('.'), {'precedence':p_value})
    def match_any(self, src):
        return ANY(src=src)

    @GProduction(regex, ARG(SET('()*+?!|{:[.\\', inv=True)),
                 {'precedence':p_value})
    def match_text(self, v, src):
        return TEXT(v, src=src)

    @GProduction(symbol, ARG(SEQ(ARG(SET([chr(x) for x in range(48,58)] +
                                         [chr(x) for x in range(65,91)] +
                                         [chr(x) for x in range(97,123)]+
                                         ['_'])), min=1)))
    def match_name(self, x, src):
        return ''.join(c for c in x)

    @GProduction(srange, CAT(ARG(NTSYM(relem)),
                            OPT(CAT(TEXT('-'), NAMED(NTSYM(relem),'b')))))
    def match_range(self, a, b=None, src=None):
        if b is None: return { a }
        return set(chr(x) for x in range(ord(a), ord(b)+1))

    @GProduction(relem, CAT(TEXT('\\'), ARG(ANY())))
    def match_rescape(self, x, src):
        return self.ESCAPE.get(x, x)

    _ = GProduction(relem, ARG(SET({'\\', ']'}, inv=True)),
                    {'action':'COPY'})

    def init_parse(self, grammar_ns):
        self.grammar = grammar_ns

    def get_symbol(self, s):
        return getattr(self.grammar, s)

PATTERN_PARSE[:] = [None, None, None]

RegexBootParser = LALRParserGenerator(RegexBootGrammar, 'RegexBootParser',
                                      SourceTrace).parser()
PATTERN_PARSE[0] = RegexBootParser

## Boot Extended BNF parser (using RegexBootParser)

class EBNFBootTokens (Grammar):
    SYMBOL = Symbol()
    LITERAL = Symbol()
    PRODUCES = Symbol()

class EBNFBootLexicalGrammar (EBNFBootTokens):

    token = NonTerminal()
    litch = NonTerminal()
    eof = EOF

    START = token

    _ = r(r'token -> \ +', action='SPACE')

    @r('token -> ->')
    def produces(self, t, src):
        return Token('->', self.PRODUCES, src=src)

    @r('token -> [*+?!|():.,]')
    def operator(self, t, src):
        return Token(t, src=src)

    @r('token -> [A-Za-z0-9_]+')
    def match_symbol(self, t, src):
         return Token(t, self.SYMBOL, src=src)

    @r(r"token -> '{litch}+'")
    def match_literal(self, t, src):
        return Token(t, self.LITERAL, ''.join(self.v), src=src)

    @r(r"litch -> [^'\\]|\\.")
    def literal_char(self, t, src):
        self.v.append(t[-1])

    def init_token(self):
        self.v = []

EBNFBootLexical = RegularParserGenerator(EBNFBootLexicalGrammar,
                                         'EBNFBootLexical', SourceTrace,
                                         LexicalAnalyser).parser()
PATTERN_PARSE[1] = EBNFBootLexical

class EBNFBootGrammar (EBNFBootTokens):

    production = NonTerminal()
    ebnf = NonTerminal()

    p_value = LeftAssociative()
    p_unary = LeftAssociative()
    p_cat = LeftAssociative()
    p_alt = LeftAssociative()

    START = production

    @r('production -> {SYMBOL:ls}{PRODUCES}{ebnf:rhs}')
    def match_production(self, ls, rhs, src):
        lhs = self.get_symbol(ls)
        if not isinstance(lhs, NonTerminal):
            raise GrammarError("Left hand side must be a non-terminal", ls)
        return (lhs, rhs, src)

    @r('ebnf -> {ebnf:a}\|{ebnf:b}', precedence=p_alt)
    def match_alt(self, a, b, src):
        return ALT(a, b, src=src)
    
    @r('ebnf -> {ebnf:a}{ebnf:b}', precedence=p_cat)
    def match_cat(self, a, b, src):
        return CAT(a, b, src=src)

    @r('ebnf -> {ebnf:v}\*(,{LITERAL:sep})?', precedence=p_unary)
    def match_seq0(self, v, sep=None, src=None):
        return SEQ(v, sep=(sep if sep is not None else None), min=0, src=src)

    @r('ebnf -> {ebnf:v}\+(,{LITERAL:sep})?', precedence=p_unary)
    def match_seq1(self, v, sep=None, src=None):
        return SEQ(v, sep=(sep if sep is not None else None), min=1, src=src)

    @r('ebnf -> {ebnf:v}\?', precedence=p_unary)
    def match_opt(self, v, src):
        return OPT(v, src=src)

    @r('ebnf -> {ebnf:v}\!', precedence=p_unary)
    def match_arg(self, v, src):
        return ARG(v, src=src)

    @r('ebnf -> {ebnf:v}\:{SYMBOL:n}', precedence=p_unary)
    def match_named(self, v, n, src):
        return NAMED(v, n, src=src)

    _ = r('ebnf -> \({ebnf:x}\)', precedence=p_value, action='COPY')

    @r('ebnf -> {SYMBOL:c}', precedence=p_value)
    def match_symbol(self, c, src):
        v = self.get_symbol(c)
        if isinstance(v, NonTerminal):
            return NTSYM(v, src=src)
        elif isinstance(v, Symbol):
            return TSYM(v, src=src)
        else:
            return TYP(c, src=src)

    @r('ebnf -> {LITERAL:v}', precedence=p_value)
    def match_text(self, v, src):
        return TEXT(v, src=src)

    def init_parse(self, grammar_ns):
        self.grammar = grammar_ns

    def get_symbol(self, s):
        return getattr(self.grammar, s)

EBNFBootParser = LALRParserGenerator(EBNFBootGrammar, 'EBNFBootParser',
                                     SourceTrace).parser()
PATTERN_PARSE[2] = EBNFBootParser

## Now build real parsers

from grammar.ptn_regex import RegexGrammar
from grammar.ptn_ebnf import EBNFLexicalGrammar, EBNFGrammar

RegexParser = LALRParserGenerator(RegexGrammar, 'RegexParser',
                                  SourceTrace).parser()
EBNFLexical = RegularParserGenerator(EBNFLexicalGrammar, "EBNFLexical",
                                     SourceTrace, LexicalAnalyser).parser()
EBNFParser = LALRParserGenerator(EBNFGrammar, "EBNFParser",
                                 SourceTrace).parser()

w = PythonWriter("grammar/ptn_tables.py")
w.add(RegexParser)
w.add(EBNFLexical)
w.add(EBNFParser)
w.write(head="""
from grammar.grammar import PATTERN_PARSE

PATTERN_PARSE[:] = [None, None, None]

""", tail="""
PATTERN_PARSE[:] = [RegexParser, EBNFLexical, EBNFParser]
""")
