
from .grammar import *
from .engine import Token

class EBNFTokens (Grammar):
    SYMBOL = Symbol()
    LITERAL = Symbol()
    PRODUCES = Symbol()

class EBNFLexicalGrammar (EBNFTokens):

    token = NonTerminal()
    litch = NonTerminal()

    START = token

    _ = r(r'token -> \ +', action='SPACE')

    _ = r('token -> ->|\!\!|[*+?!|():.,]', action='LITERAL_TOKEN')

    _ = r('token -> [A-Za-z0-9_]+', action='SYMBOL')

    @r(r"token -> '{litch}+'")
    def match_literal(self, t, src):
        return Token(t, self.LITERAL, ''.join(self.v), src=src)

    @r(r"litch -> [^'\\]|\\.")
    def literal_char(self, t, src):
        self.v.append(t[-1])

    def init_token(self):
        self.v = []

class EBNFGrammar (EBNFTokens):

    production = NonTerminal()
    ebnf = NonTerminal()

    p_value = LeftAssociative()
    p_unary = LeftAssociative()
    p_cat = LeftAssociative()
    p_alt = LeftAssociative()

    START = production

    @p("production -> SYMBOL! '->' ebnf!")
    def match_production(self, ls, rhs, src):
        lhs = self.get_symbol(ls)
        if not isinstance(lhs, NonTerminal):
            raise GrammarError("Left hand side must be a non-terminal", ls)
        return (lhs, rhs, src)

    @p("ebnf -> ebnf! '|' ebnf!", precedence=p_alt)
    def match_alt(self, a, b, src):
        return ALT(a, b, src=src)
    
    @p("ebnf -> ebnf! ebnf!", precedence=p_cat)
    def match_cat(self, a, b, src):
        return CAT(a, b, src=src)

    @p("ebnf -> ebnf! '*' ( ',' LITERAL! )?", precedence=p_unary)
    def match_seq0(self, v, sep=None, src=None):
        return SEQ(v, sep=(sep if sep is not None else None), min=0, src=src)

    @p("ebnf -> ebnf! '+' ( ',' LITERAL! )?", precedence=p_unary)
    def match_seq1(self, v, sep=None, src=None):
        return SEQ(v, sep=(sep if sep is not None else None), min=1, src=src)

    @p("ebnf -> ebnf! '?'", precedence=p_unary)
    def match_opt(self, v, src):
        return OPT(v, src=src)

    @p("ebnf -> ebnf! '!'", precedence=p_unary)
    def match_arg(self, v, src):
        return ARG(v, src=src)

    @p("ebnf -> ebnf! '!!'", precedence=p_unary)
    def match_dictarg(self, v, src):
        return DICTARG(v, src=src)

    @p("ebnf -> ebnf! ':' SYMBOL!", precedence=p_unary)
    def match_named(self, v, n, src):
        return NAMED(v, n, src=src)

    _ = p("ebnf -> '(' ebnf! ')'", precedence=p_value, action='COPY')

    @p("ebnf -> SYMBOL!", precedence=p_value)
    def match_symbol(self, c, src):
        v = self.get_symbol(c)
        if isinstance(v, NonTerminal):
            return NTSYM(v, src=src)
        elif isinstance(v, Symbol):
            return TSYM(v, src=src)
        elif isinstance(v, Category):
            return TYP(c, src=src)
        elif isinstance(v, RMacro):
            return v.compile(self.grammar)
        else:
            raise GrammarError("Reference to object of unrecognised type", src)

    @p("ebnf -> LITERAL!", precedence=p_value)
    def match_text(self, v, src):
        return TEXT(v, src=src)

    def init_parse(self, grammar_ns):
        self.grammar = grammar_ns

    def get_symbol(self, s):
        return getattr(self.grammar, s)
