
from functools import reduce

from .grammar import *

class RegexGrammar (Grammar):
    production = NonTerminal()
    regex = NonTerminal()
    symbol = NonTerminal()
    srange = NonTerminal()
    relem = NonTerminal()

    p_value = LeftAssociative()
    p_unary = LeftAssociative()
    p_cat = LeftAssociative()
    p_alt = LeftAssociative()

    START = (production, regex)
    ESCAPE = { 'n': '\n', 'r': '\n', 't': '\t' }

    @p(r"production -> symbol! ' '* '-' '>' ' '* regex!")
    def match_production(self, ls, rhs, src):
        lhs = self.get_symbol(ls)
        if not isinstance(lhs, NonTerminal):
            raise GrammarError("Left hand side must be a non-terminal", ls)
        return (lhs, rhs, src)

    @p(r"regex -> regex! '|' regex!", precedence=p_alt)
    def match_alt(self, a, b, src):
        return ALT(a, b, src=src)

    @p(r"regex -> regex! regex!", precedence=p_cat)
    def match_cat(self, a, b, src):
        return CAT(a, b, src=src)

    @p(r"regex -> regex! '*'", precedence=p_unary)
    def match_seq0(self, v, src):
        return SEQ(v, min=0, src=src)

    @p(r"regex -> regex! '+'", precedence=p_unary)
    def match_seq1(self, v, src):
        return SEQ(v, min=1, src=src)

    @p(r"regex -> regex! '?'", precedence=p_unary)
    def match_opt(self, v, src):
        return OPT(v, src=src)

    @p(r"regex -> regex! '!'", precedence=p_unary)
    def match_arg(self, v, src):
        return ARG(v, src=src)

    @p(r"regex -> regex! ':' symbol!", precedence=p_unary)
    def match_named(self, v, n, src):
        return NAMED(v, n, src=src)

    _ = p(r"regex -> '(' regex! ')'", precedence=p_value, action='COPY')

    @p(r"regex -> '[' '^'?:inv srange!+! ']'", precedence=p_value)
    def match_set(self, s, inv, src):
        return SET(reduce(set.union, s), inv, src=src)

    @p(r"regex -> '{' symbol! (':' symbol!)? '}'", precedence=p_value)
    def match_symbol(self, c, n=None, src=None):
        v = self.get_symbol(c)
        if isinstance(v, NonTerminal):
            o = NTSYM(v, src=src)
        elif isinstance(v, Symbol):
            o = TSYM(v, src=src)
        elif isinstance(v, Category):
            o = TYP(c, src=src)
        elif isinstance(v, RMacro):
            o = v.compile(self.grammar)
        else:
            raise GrammarError("Reference to object of unrecognised type", src)
        if n is None: return o
        return NAMED(o,n, src)

    @r(r"regex -> \\.!", precedence=p_value)
    def match_escape(self, v, src):
        return TEXT(self.ESCAPE.get(v, v), src=src)

    @p(r"regex -> '.'", precedence=p_value)
    def match_any(self, src):
        return ANY(src=src)

    @r(r"regex -> [^()*+?!|{:[.\\]!", precedence=p_value)
    def match_text(self, v, src):
        return TEXT(v, src=src)

    @r(r"symbol -> [A-Za-z0-9_]!+!")
    def match_name(self, x, src):
        return ''.join(c for c in x)

    @p(r"srange -> relem:a ('-' relem:b)?")
    def match_range(self, a, b=None, src=None):
        if b is None: return { a }
        return set(chr(x) for x in range(ord(a), ord(b)+1))

    @r(r"relem -> \\.!")
    def match_rescape(self, x, src):
        return self.ESCAPE.get(x, x)

    _ = r(r"relem -> [^\\\]]!", action='COPY')

    def init_parse(self, grammar_ns):
        self.grammar = grammar_ns

    def get_symbol(self, s):
        return getattr(self.grammar, s)
