
class frozendict(dict):
    "Taken from http://code.activestate.com/recipes/414283-frozen-dictionaries/"
    @property
    def _blocked_attribute(obj):
        raise AttributeError("A frozendict cannot be modified.")

    __delitem__ = __setitem__ = clear = _blocked_attribute
    pop = popitem = setdefault = update = _blocked_attribute

    def __new__(cls, *args):
        new = dict.__new__(cls)
        dict.__init__(new, *args)
        return new

    def __init__(self, *args):
        pass

    def __hash__(self):
        try:
            return self._cached_hash
        except AttributeError:
            h = self._cached_hash = hash(tuple(sorted(self.items())))
            return h

    def __repr__(self):
        return "frozendict({})".format(dict.__repr__(self))

class TopologicalLoop (Exception):
    pass

def topol(c):
    """Topological attribute propogation.

    c is a dict, mapping nodes to a pair of sets, the first a set of
    native attributes, the second a set of nodes to inherit attributes
    from (ie the graph edges).

    The result will be a dict mapping nodes to a set of attributes,
    combining native and inherited attributes for the node.
    """
    from functools import reduce
    def union(a, b):
        return (a[0] | b[0], a[1] + b[1])
    def visit(n, r, c):
        a = r.get(n)
        if a == 1:
            return (set(), [[n,True,{n}]])
        elif a == 2:
            raise TopologicalLoop(n)
        elif a is not None:
            return (a, [])
        r[n] = 1
        na,ia = c.get(n, (set(), set()))
        a,lpl = reduce(union, (visit(x, r, c) for x in ia), (set(), []))
        a |= na
        lpo = True
        for lp in lpl:
            if lp[1]:
                if lp[0] == n:
                    lp[1] = False
                    lpo = False
                else:
                    lp[2].add(n)
            else:
                lpo = False
        if not lpo:
            for nn in reduce(set.union, (lp[2] for lp in lpl), set()):
                r[nn] = a
            return (a, [])
        elif lpl:
            r[n] = 2
            return (a, lpl)
        r[n] = a
        return (a, lpl)
    r = {}
    for n in c:
        visit(n, r, c)
    return r

def find_loops(c, report):
    "Find topological loops"
    
    from functools import reduce
    def union(a, b):
        return (a[0] | b[0], a[1] + b[1])
    def visit(n, r, c):
        a = r.get(n)
        if a == 1:
            return (set(), [[n,True,{n}]])
        elif a == 2:
            return (set(), [])
        elif a is not None:
            return (a, [])
        r[n] = 1
        na,ia = c.get(n, (set(), set()))
        a,lpl = reduce(union, (visit(x, r, c) for x in ia), (set(), []))
        a |= na
        lpo = True
        for lp in lpl:
            if lp[1]:
                if lp[0] == n:
                    lp[1] = False
                    lpo = False
                    report(lp)
                else:
                    lp[2].add(n)
            else:
                lpo = False
        if not lpo:
            for nn in reduce(set.union, (lp[2] for lp in lpl), set()):
                r[nn] = a
            return (a, [])
        elif lpl:
            r[n] = 2
            return (a, lpl)
        r[n] = a
        return (a, lpl)
    r = {}
    for n in c:
        visit(n, r, c)



class Topology:
    def __init__(self, lbl):
        self.lbl = lbl
        self.unfinished = 1
        self.loops = []
        self.deps = set()
        self.attrs = set()

    def visit(self, n, r, c):
        from functools import reduce
        t = r.get(n)
        if t is not None:
            if t.unfinished:
                self.loops.append([n,t.unfinished,{n}])
            else:
                self.attrs |= t.attrs
                self.deps |= t.deps
                self.deps.add(n)
            return
        t = Topology(n)
        r[n] = t
        na,ia = c.get(n, (set(), set()))
        t.attrs |= na
        for ni in ia:
            t.visit(ni, r, c)
        over = True
        for loop in t.loops:
            if loop[1] == 2:
                if loop[0] in t.deps:
                    loop[1] = 0
                else:
                    loop[2].add(n)
                    over = False
            elif loop[1] == 1:
                if loop[0] == n:
                    loop[1] = 0
                else:
                    loop[2].add(n)
                    over = False
        if over and t.loops:
            #print("Loop :", reduce(set.union, (loop[2] for loop in t.loops)))
            for nn in reduce(set.union, (loop[2] for loop in t.loops)):
                r[nn].attrs |= t.attrs
                r[nn].unfinished = 0
            t.unfinished = 0
        elif t.loops:
            self.loops += t.loops
            t.unfinished = 2
        else:
            t.unfinished = 0
        self.attrs |= t.attrs
        self.deps |= t.deps
        self.deps.add(n)

def new_topol(c):
    """Topological attribute propogation.

    c is a dict, mapping nodes to a pair of sets, the first a set of
    native attributes, the second a set of nodes to inherit attributes
    from (ie the graph edges).

    The result will be a dict mapping nodes to a set of attributes,
    combining native and inherited attributes for the node.
    """
    t = Topology(None)
    r = {}
    for n in c:
        t.visit(n, r, c)
    return { n:r[n].attrs for n in c }

class ExtendsType (type):
    def __new__(mcls, name, bases, attrs):
        if not bases: return super().__new__(mcls, name, bases, attrs)
        from sys import modules
        m = modules[attrs.pop('__module__')]
        if hasattr(bases[0], 'SRCMODULE'): m = bases[0].SRCMODULE
        cls = getattr(m, name)
        for k,v in attrs.items():
            setattr(cls, k, v)
        return cls

class Extension (metaclass=ExtendsType):
    pass

class EnumBaseClass (type):
    def __new__(metacls, name, parents, attrs):
        vlist = attrs.get("VALUES", [])
        cls = super().__new__(metacls, name, parents, attrs)
        for i,v in enumerate(vlist):
            setattr(cls, v, cls(i))
        return cls

class enum (int, metaclass=EnumBaseClass):
    def __repr__(self):
        return type(self).__name__+'.'+self.VALUES[self]
